/* $AoiMoe: xgesture.cc,v 1.7 2003/02/06 14:31:00 tshiozak Exp $ */

/*-
 * Copyright (c)2003 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <iostream>
#include <string>
#include <list>
#include <map>
#include <algorithm>
#include <cmath>
#include <exception>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include <X11/extensions/XTest.h>
#include <X11/Xmu/WinUtil.h>

#include "Grabber.h"

#include "gesture_conf.hh"

using namespace std;

#if 0
#define DPRCODE(x) x
#else
#define DPRCODE(x)
#endif

typedef struct {
	String up_char, down_char, left_char, right_char, gestures;
	String paren_left, paren_right, square_left, square_right;
	int unit_distance, max_strokes;
	unsigned int button, modifiers;
	Boolean warp_back;
} AppResRec, *AppRes;

static XtResource app_resources[] = {
	{ "upChar", "UpChar", XtRString, sizeof(String),
	  XtOffset(AppRes, up_char), XtRString, XtPointer("u") },
	{ "downChar", "DownChar", XtRString, sizeof(String),
	  XtOffset(AppRes, down_char), XtRString, XtPointer("d") },
	{ "leftChar", "LeftChar", XtRString, sizeof(String),
	  XtOffset(AppRes, left_char), XtRString, XtPointer("l") },
	{ "rightChar", "RightChar", XtRString, sizeof(String),
	  XtOffset(AppRes, right_char), XtRString, XtPointer("r") },
	{ "unitDistance", "UnitDistance", XtRInt, sizeof(int),
	  XtOffset(AppRes, unit_distance), XtRImmediate, XtPointer(10) },
	{ "maxStrokes", "MaxStrokes", XtRInt, sizeof(int),
	  XtOffset(AppRes, max_strokes), XtRImmediate, XtPointer(10) },
	{ "button", "Button", XtRInt, sizeof(int),
	  XtOffset(AppRes, button), XtRImmediate, XtPointer(3) },
	{ "modifiers", "Modifiers", XtRInt, sizeof(int),
	  XtOffset(AppRes, modifiers), XtRImmediate, XtPointer(0) },
	{ "warpBack", "WarpBack", XtRBoolean, sizeof(Boolean),
	  XtOffset(AppRes, warp_back), XtRString, XtPointer("True") },
	{ "gestures", "Gestures", XtRString, sizeof(String),
	  XtOffset(AppRes, gestures), XtRString, XtPointer("") },
	{ "parenLeft", "ParenLeft", XtRString, sizeof(String),
	  XtOffset(AppRes, paren_left), XtRString, XtPointer("(") },
	{ "parenRight", "ParenRight", XtRString, sizeof(String),
	  XtOffset(AppRes, paren_right), XtRString, XtPointer(")") },
	{ "squareLeft", "SquareLeft", XtRString, sizeof(String),
	  XtOffset(AppRes, square_left), XtRString, XtPointer("[") },
	{ "squareRight", "SquareRight", XtRString, sizeof(String),
	  XtOffset(AppRes, square_right), XtRString, XtPointer("]") },
};

String fallback_resources[] = {
	"*grabber.translations: #override\\n"
	"	<Motion>: Grabber()\\n"
	"	<BtnDown>: Grabber()\\n"
	"	<BtnUp>: Grabber()\\n"
	"",
	"*label.label: motion:",
	"*Label.justify: left",
	"*Label.borderWidth: 0",
	NULL,
};

#define ACTION_PROTO(name) \
static void name(Widget w, XEvent *event, String *params, Cardinal *nparams)

ACTION_PROTO(Grabber);
XtActionsRec actions[] = {
	{ "Grabber", &Grabber },
};


Display *dpy;
AppResRec appres;
Widget shell, grabber, appname, label, status;
Window curwin;
int prev_x, prev_y, begin_x, begin_y;
typedef enum {
	DIR_NONE, DIR_R, DIR_L, DIR_U, DIR_D
} Direction;
Direction prev_dir;
int strokes;
string motion_string, motion_print, client_name, matched_string;
string wmresclass, wmresname, wmname;
unsigned int grabbed_button, grabbed_modifier;


//

class XKeysymConverter_C : public KeysymConverter_I
{
	Display *m_display_;
public:
	XKeysymConverter_C(Display *d) : m_display_(d) { }
	~XKeysymConverter_C() { }
	int operator () (string const &keysym)
	{
		int keycode;

		keycode = XKeysymToKeycode(m_display_,
					   XStringToKeysym(keysym.c_str()));
		if (keycode==0)
			throw UnknownKeysym_E(keysym);
		return (keycode);
	}
};

class ErrorMessageProc_C : public ErrorMessageProc_I
{
public:
	ErrorMessageProc_C() { }
	~ErrorMessageProc_C() { }
	void operator () (int line, string const &err, string const &token)
	{
		cerr << line << ": " << err << ", near \"" << token << "\""
		     << endl;
	}
};


class GenKey_C
{
	Display *m_display_;
public:
	GenKey_C(Display *d) : m_display_(d) { }
	void operator () (int keycode, bool ispress)
	{
		XTestFakeKeyEvent(m_display_, keycode, ispress, 10);
		XSync(m_display_, False);
	}
};


class TargetMatch_C
{
	std::string m_wmname_, m_resname_, m_resclass_;
public:
	TargetMatch_C(std::string const &wmn,
		      std::string const &rn,
		      std::string const &rc)
		: m_wmname_(wmn), m_resname_(rn), m_resclass_(rc)
	{
	}
	bool operator () (Target_C::Handle ml)
	{
		return (ml->GetAppName() == m_wmname_ ||
			ml->GetAppName() == m_resname_ ||
			ml->GetAppName() == m_resclass_);
	}
};


static ParseManager_C::Handle parseManager;
static TargetPool_C::Handle targetPool;
static Target_C::Handle currentTarget;
static Motion_C::Handle currentMotion;

void
ParseGestures()
{
	parseManager = new ParseManager_C(new XKeysymConverter_C(dpy),
					  new ErrorMessageProc_C); // XXX
	parseManager->Parse(appres.gestures);
	targetPool = parseManager->GetTargetPool();
}

static void
UpdateStatusWindow()
{
	string tmp;
	list<Widget> widgets;
	Dimension totalw, totalh, w, h ,bw;

	if (client_name.length() > 0) {
		tmp  = appres.square_left;
		tmp += client_name;
		tmp += appres.square_right;
	}
	XtVaSetValues(appname, XtNlabel, tmp.c_str(), NULL);

	tmp = motion_print;
	if (matched_string.length() > 0) {
		tmp += appres.paren_left;
		tmp += matched_string;
		tmp += appres.paren_right;
	}
	XtVaSetValues(status, XtNlabel, tmp.c_str(), NULL);

	widgets.push_back(appname);
	widgets.push_back(label);
	widgets.push_back(status);

	totalw = totalh = 0;
	for (list<Widget>::iterator i=widgets.begin(); i!=widgets.end(); ++i) {
		XtVaSetValues(*i, XtNx, totalw, NULL);
		XtVaGetValues(*i,
			      XtNwidth, &w,
			      XtNheight, &h,
			      XtNborderWidth, &bw,
			      NULL);
		totalw += w+bw*2;
		totalh = max<Dimension>(totalh, h+bw*2);
	}
	XtVaSetValues(shell, XtNwidth, totalw, XtNheight, totalh, NULL);
}

static void
ConvertGrabbedButton()
{
	grabbed_button = appres.button; // XXX
	grabbed_modifier = appres.modifiers; // XXX
}

static void
SetPassiveGrab()
{
	XtGrabButton(grabber, grabbed_button, grabbed_modifier, False,
		     (PointerMotionMask|ButtonPressMask|ButtonReleaseMask|
		     OwnerGrabButtonMask),
		     GrabModeAsync, GrabModeAsync, None, None);
	XSync(dpy, False);
}

static void
UnsetPassiveGrab()
{
	XtUngrabButton(grabber, grabbed_button, grabbed_modifier);
	XSync(dpy, False);
}

static void
ProcMotion(XMotionEvent *ev)
{
	int dis_x, dis_y;
	Direction dir;
	String pch;
	char ch;

	if (!currentTarget || strokes >= appres.max_strokes)
		return;

	dis_x = abs(ev->x_root - prev_x);
	dis_y = abs(ev->y_root - prev_y);

	if (dis_x < appres.unit_distance && dis_y < appres.unit_distance)
		return;

	if (dis_x==dis_y)
		return;
	if (dis_x>dis_y) {
		if (ev->x_root > prev_x)
			dir = DIR_R;
		else
			dir = DIR_L;
	} else {
		if (ev->y_root > prev_y)
			dir = DIR_D;
		else
			dir = DIR_U;
	}
	if (dir != prev_dir) {
		switch (dir) {
		case DIR_R:
			pch = appres.right_char;
			ch = 'r';
			break;
		case DIR_L:
			pch = appres.left_char;
			ch = 'l';
			break;
		case DIR_U:
			pch = appres.up_char;
			ch = 'u';
			break;
		case DIR_D:
			pch = appres.down_char;
			ch = 'd';
			break;
		}
		motion_print += pch;
		motion_string += ch;
		string().swap(matched_string);
		currentMotion = NULL;
		if (currentTarget) {
			MotionSet_C::Handle ml =
			    currentTarget->GetMotionSet();
			currentMotion = ml->GetMotion(motion_string);
			if (currentMotion) {
				matched_string = currentMotion->GetTitle();
			}
		}
		UpdateStatusWindow();
		if (strokes==0) {
			XtMapWidget(shell);
			XRaiseWindow(dpy, XtWindow(shell));
		}
		strokes++;
		DPRCODE(cerr << ch);
		DPRCODE(cerr.flush());
	}
	prev_dir = dir;
	prev_x = ev->x_root;
	prev_y = ev->y_root;
}

static void
ProcButton(XButtonEvent *ev)
{
	int dummy_int;
	unsigned int dummy_uint;
	Window dummy_win, onwin;

	switch (ev->type) {
	case ButtonPress:
		DPRCODE(cerr << "button down" << endl);
		begin_x = prev_x = ev->x_root;
		begin_y = prev_y = ev->y_root;
		prev_dir = DIR_NONE;
		strokes = 0;
		string().swap(motion_string);
		string().swap(motion_print);
		string().swap(matched_string);
		string().swap(client_name);
		string().swap(wmname);
		string().swap(wmresname);
		string().swap(wmresclass);
		XQueryPointer(dpy, XtWindow(grabber),
			      &dummy_win, &onwin,
			      &dummy_int, &dummy_int, &dummy_int, &dummy_int,
			      &dummy_uint);
		if (onwin != None)
			curwin = XmuClientWindow(dpy, onwin);
		if (curwin != None) {
			XClassHint *chint;
			char *name;
			DPRCODE(cerr << "current client win: "
				<< (void *)curwin << endl);
			chint = XAllocClassHint();
			XGetClassHint(dpy, curwin, chint);
			if (chint->res_name)
				wmresname = chint->res_name;
			if (chint->res_class)
				wmresclass = chint->res_class;
			XFree(chint->res_name);
			XFree(chint->res_class);
			XFree(chint);
			XFetchName(dpy, curwin, &name);
			if (name)
				wmname = name;
			XFree(name);
			DPRCODE(cerr << "\tresname=" << wmresname
				<< ", resclass=" << wmresclass
				<< ", name=" << name << endl);
		}
		try {
			currentTarget =
			    targetPool->Lookup(TargetMatch_C(wmname,
							     wmresname,
							     wmresclass));
			client_name = currentTarget->GetTitle();
		}
		catch (NoEntry_E &) {
			currentTarget = NULL;
		}
		break;
	case ButtonRelease:
		DPRCODE(cerr << endl << "button up" << endl);
		if (appres.warp_back) {
			XWarpPointer(dpy, None, XtWindow(grabber),
				     0, 0, 0, 0, begin_x, begin_y);
			XSync(dpy, False);
		}
		if (prev_dir == DIR_NONE) {
			if (curwin != None) {
				UnsetPassiveGrab();
				XTestFakeButtonEvent(dpy,
						     grabbed_button,
						     True,
						     0);
				SetPassiveGrab();
				XTestFakeButtonEvent(dpy,
						     grabbed_button,
						     False,
						     10);
			}
		} else {
			if (currentMotion) {
				currentMotion->GetSequence()->GenerateKey(
					GenKey_C(dpy));
			}
			curwin = None;
			XtUnmapWidget(shell);
		}
		currentMotion = NULL;
		currentTarget = NULL;
		break;
	}
}

ACTION_PROTO(Grabber)
{
	switch (event->type) {
	case MotionNotify:
		ProcMotion(&event->xmotion);
		break;
	case ButtonPress:
	case ButtonRelease:
		ProcButton(&event->xbutton);
		break;
	default:
		DPRCODE(cerr << "default" << endl);
		break;
	}
}

int
main(int argc, char **argv)
{
	XtAppContext appcon;
	int dum;

	XtSetLanguageProc(NULL, NULL, NULL);

	shell = XtVaOpenApplication(&appcon, "XGesture", NULL, 0,
				    &argc, argv, fallback_resources,
				    overrideShellWidgetClass,
				    XtNwidth, 1, XtNheight, 1,
				    NULL);

	dpy = XtDisplay(shell);

	if (!XTestQueryExtension(dpy, &dum, &dum, &dum, &dum)) {
		cerr << "X Test extension is not supported." << endl;
		return (1);
	}

	XtVaGetApplicationResources(shell, XtPointer(&appres),
				    app_resources, XtNumber(app_resources),
				    NULL);

	ConvertGrabbedButton();
	ParseGestures();

	appname = XtVaCreateWidget("appname", labelWidgetClass, shell, NULL);
	label = XtVaCreateWidget("label", labelWidgetClass, shell, NULL);
	status = XtVaCreateWidget("status", labelWidgetClass, shell,
				  XtNlabel, "",
				  NULL);
	grabber = XtVaCreateWidget("grabber", grabberWidgetClass,
				   shell, NULL);

	XtAppAddActions(appcon, actions, XtNumber(actions));

	XtSetMappedWhenManaged(shell, False);
	XtRealizeWidget(shell);
	XtRealizeWidget(label);
	XtRealizeWidget(status);
	XtRealizeWidget(grabber);
	XtMapWidget(appname);
	XtMapWidget(label);
	XtMapWidget(status);

	SetPassiveGrab();

	XtAppMainLoop(appcon);
	/* NOTREACHED */
}
