/* $AoiMoe: yacc.y,v 1.5 2003/02/06 13:34:17 tshiozak Exp $ */

/*-
 * Copyright (c)2003 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

%{
#include <stdio.h>
#include "gesture_conf_p.h"
%}

%union {
	char *str;
	int i;
}
%token			T_DEFM
%token			T_TARGET
%token	<str>		T_STRING
%token	<str>		T_KEYSYM
%token	<str>		T_MOTION
%token	<str>		T_ATOM
%token	<str>		T_GARBAGE

%type	<str>	title

%%

gestures	: /* */
		| gestures def
		;

def		: def_motions
		| def_target
		;

def_motions	: T_DEFM T_ATOM '{' motion_set '}'
		{ CreateNamedMotionSetFromStack($2); }

motion_set	: /* */
		| motion_set import_motionset
		| motion_set motion

import_motionset: '<' T_ATOM '>' ';'
		{ ImportMotion($2); }

motion		: T_MOTION title '[' strokes ']' ';'
		{ PushSequence(); PushNewMotion($1, $2); }

title		: /* */
		{ $$ = NULL; }
		| T_STRING

strokes		: /* */
		| strokes stroke
		| strokes release_mark

stroke		: T_KEYSYM
		{ PushStroke($1); }

release_mark	: ';'
		{ PushSequence(); }


def_target	: T_TARGET T_STRING title '{' motion_set '}' ';'
		{ CreateNewTargetFromStack($2, $3); }
		| T_TARGET T_STRING title T_ATOM ';'
		{ CreateNewTargetFromNamedMotionSet($2, $3, $4); }

%%

int
yyerror(const char *s)
{
	char buf[2];
	switch (yychar) {
	case T_DEFM:
		ProcGesturesError(s, "DEFM");
		break;
	case T_TARGET:
		ProcGesturesError(s, "TARGET");
		break;
	case T_STRING:
	case T_ATOM:
	case T_KEYSYM:
	case T_MOTION:
	case T_GARBAGE:
		ProcGesturesError(s, yylval.str);
		break;
	default:
		buf[0] = yychar;
		buf[1] = 0;
		ProcGesturesError(s, buf);
		break;
	}
	return (0);
}
